"""
emulation TD4
"""

import td4.cpu
import td4.output
import td4.opcode
import td4.parse

import argparse
import time


def emulator():
    # Initialize
    pc = 0  # Program Counter
    reg_a = "0000"  # Register A
    reg_b = "0000"  # Register B
    c_flag = "0"  # Carry Flag
    output = "0000"  # Output

    # Parse arguments
    args = arg_parse()
    file = args.file
    input = args.input
    clk = args.clock
    beep = args.beep

    # Check input
    try:
        if len(input) != 4 or not input.isnumeric() or int(input, 2) > 15:
            raise ValueError
    except ValueError:
        print("Invalid input")
        exit(1)

    # Read file
    try:
        with open(file, "r") as file:
            if file.name.endswith(".td4"):
                rom, clk, beep = td4.parse.to_binary_td4(file.read().splitlines())
            else:
                rom = [td4.parse.to_binary(line) for line in file.read().splitlines()]
        if len(rom) > 16:
            raise ValueError
    except FileNotFoundError:
        print("File not found")
        exit(1)
    except ValueError:
        print("Too many instructions\nTD4 can only handle 16 instructions")
        exit(1)
    except KeyError:
        print("Invalid instruction")
        exit(1)

    # Emulate
    try:
        td4.output.output(reg_a, reg_b, c_flag, rom, pc, clk, input, output, beep)
        while pc < len(rom):
            inst, pc = td4.cpu.fetch(rom, pc)
            op, im = td4.cpu.decode(inst)
            reg_a, reg_b, c_flag, output, pc = td4.cpu.execute(
                op, im, reg_a, reg_b, c_flag, input, output, pc
            )
            if beep and output[0] == "1":
                print("\a")
            clock(clk)
            td4.output.output(reg_a, reg_b, c_flag, rom, pc, clk, input, output, beep)
    except KeyboardInterrupt:
        print("Finished")
        exit(0)
    except KeyError:
        print("Invalid instruction")
        exit(1)


def clock(clk):
    if clk.isnumeric():
        interval = 1 / int(clk)
        time.sleep(interval)
    elif clk == "manual":
        input()


def arg_parse():
    url = "https://github.com/yashikota/td4-py#support-file-format"
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        help=f"File to read\nCheck {url} for the supported format.",
        type=str,
    )
    parser.add_argument(
        "-i",
        "--input",
        help="Default 0000. Any binary can be specified.",
        type=str,
        default="0000",
    )
    parser.add_argument(
        "-c",
        "--clock",
        help="Default 10Hz. Any number and manual can be specified.",
        type=str,
        default="10",
    )
    parser.add_argument("-b", "--beep", help="Default False.", action="store_true")
    args = parser.parse_args()

    return args



"""
Parse Input
"""

import re
import td4.opcode


def to_binary(i):
    instruction = re.split(r"[, ]", i)

    if len(instruction) == 3 and instruction[2].isalpha():
        # MOV A, B | MOV B, A
        return (
            td4.opcode.to_opcode("".join(list(map(lambda x: x.upper(), instruction))))
            + "0000"
        )
    elif len(instruction) == 3 and instruction[1].isalpha():
        # MOV A, Im | MOV B, Im | ADD A, Im | ADD B, Im
        return (
            td4.opcode.to_opcode(
                "".join(list(map(lambda x: x.upper(), instruction[:2])))
            )
            + instruction[2]
        )
    elif len(instruction) == 2 and instruction[1].isalpha():
        # IN A | IN B | OUT B
        return (
            td4.opcode.to_opcode("".join(list(map(lambda x: x.upper(), instruction))))
            + "0000"
        )
    elif len(instruction) == 2 and instruction[0].isalpha():
        # OUT Im | JMP Im | JNC Im
        return td4.opcode.to_opcode(instruction[0].upper()) + instruction[1]
    else:
        # Not Assembly
        return "".join(instruction)


def to_binary_td4(file):
    INSTRUCTION_LENGTH = 8

    instruction = [
        ("".join(file[index : index + INSTRUCTION_LENGTH]))[::-1]
        for index in range(0, 128, INSTRUCTION_LENGTH)
    ]
    if file[128] == "#TRUE#":
        clock = "1"
    elif file[129] == "#TRUE#":
        clock = "10"
    else:
        clock = "manual"
    beep = True if file[131] == "1" else False

    return instruction, clock, beep


def to_assembly(instruction):
    return td4.opcode.to_assembly(instruction[:4], instruction[4:])



def fetch(rom, pc):
    return rom[pc], pc + 1


def decode(inst):
    return inst[:4], inst[4:]


def execute(op, im, reg_a, reg_b, c_flag, input, output, pc):
    match op:
        case "0011":  # MOV A, Im
            reg_a = im
            c_flag = "0"
        case "0111":  # MOV B, Im
            reg_b = im
            c_flag = "0"
        case "0001":  # MOV A, B
            reg_a = reg_b
            c_flag = "0"
        case "0100":  # MOV B, A
            reg_b = reg_a
            c_flag = "0"
        case "0000":  # ADD A, Im
            reg_a, c_flag = add(reg_a, im)
        case "0101":  # ADD B, Im
            reg_b, c_flag = add(reg_b, im)
        case "0010":  # IN A
            reg_a = input
            c_flag = "0"
        case "0110":  # IN B
            reg_b = input
            c_flag = "0"
        case "1011":  # OUT Im
            output = im
            c_flag = "0"
        case "1001":  # OUT B
            output = reg_b
            c_flag = "0"
        case "1111":  # JMP
            pc = int(im, 2)
            c_flag = "0"
        case "1110":  # JNC
            if c_flag == "0":
                pc = int(im, 2)

    return reg_a, reg_b, c_flag, output, pc


def add(reg, im):
    result = bin(int(reg, 2) + int(im, 2))[2:].zfill(4)
    if len(result) > 4:
        c_flag = "1"
        result = result[1:]
    else:
        c_flag = "0"
    return result, c_flag



import td4.parse


def output(reg_a, reg_b, c_flag, im, pc, clock, input, output, beep):
    print("\033[2J\033[1;1H", end="")

    print("======= TD4 Emulator =======")
    print(f"Reg A\t: {reg_a}")
    print(f"Reg B\t: {reg_b}")
    print(f"C Flag\t: {c_flag}")
    print(f"PC\t: {pc} ({bin(pc)[2:].zfill(4)})")
    print()
    if clock.isnumeric():
        print(f"Clock\t: {clock}Hz")
    elif clock == "manual":
        print("Clock\t: manual")
    print(f"Input\t: {input}")
    print(f"Output\t: {output} {led(output)}")
    print(f"Beep\t: {beep}")
    print("=========== ROM ============")
    for i, inst in enumerate(im):
        if i == pc:
            print(
                "\033[1;31m[{0:2d}] {1:<11} ({2} {3})\033[0m".format(
                    i, td4.parse.to_assembly(inst), inst[:4], inst[4:]
                )
            )
        else:
            print(
                "[{0:2d}] {1:<11} ({2} {3})".format(
                    i, td4.parse.to_assembly(inst), inst[:4], inst[4:]
                )
            )
    print("============================")


def led(output):
    return "".join([" ●" if output[i] == "1" else " ○" for i in range(4)])
